@@ -10,8 +10,10 @@ from django.core.files import File
 from django.core.files.storage import default_storage
 from django.template import Context, Template
 from django.utils import timezone
+from django.utils.module_loading import import_string
 from webob import Response
 import pkg_resources
+from six import string_types
 
 from web_fragments.fragment import Fragment
 from xblock.core import XBlock
@@ -40,9 +42,29 @@ class ScormXBlock(XBlock):
         XBLOCK_SETTINGS["ScormXBlock"] = {
             "LOCATION": "alternatevalue",
         }
-
+    
     Note that neither the folder the folder nor the package file are deleted when the
     xblock is removed.
+
+    By default, static assets are stored in the default Django storage backend. To
+    override this behaviour, you should define a custom storage function. This
+    function must take the xblock instance as its first and only argument. For instance,
+    you can store assets in different directories depending on the XBlock organisation with::
+
+        def scorm_storage(xblock):
+            from django.conf import settings
+            from django.core.files.storage import FileSystemStorage
+            from openedx.core.djangoapps.site_configuration.models import SiteConfiguration
+            
+            subfolder = SiteConfiguration.get_value_for_org(
+                xblock.location.org, "SCORM_STORAGE_NAME", "default"
+            )
+            storage_location = os.path.join(settings.MEDIA_ROOT, subfolder)
+            return get_storage_class(settings.DEFAULT_FILE_STORAGE)(location=storage_location)
+
+        XBLOCK_SETTINGS["ScormXBlock"] = {
+            "STORAGE_FUNC": scorm_storage,
+        }
     """
 
     display_name = String(
@@ -174,50 +196,91 @@ class ScormXBlock(XBlock):
         self.update_package_meta(package_file)
 
         # First, save scorm file in the storage for mobile clients
-        if default_storage.exists(self.package_path):
+        if self.storage.exists(self.package_path):
             logger.info('Removing previously uploaded "%s"', self.package_path)
-            default_storage.delete(self.package_path)
-        default_storage.save(self.package_path, File(package_file))
+            self.storage.delete(self.package_path)
+        self.storage.save(self.package_path, File(package_file))
         logger.info('Scorm "%s" file stored at "%s"', package_file, self.package_path)
 
-        # Then, extract zip file
-        if default_storage.exists(self.extract_folder_base_path):
-            logger.info(
-                'Removing previously unzipped "%s"', self.extract_folder_base_path
-            )
-            recursive_delete(self.extract_folder_base_path)
-        with zipfile.ZipFile(package_file, "r") as scorm_zipfile:
-            for zipinfo in scorm_zipfile.infolist():
-                # Do not unzip folders, only files. In Python 3.6 we will have access to
-                # the is_dir() method to verify whether a ZipInfo object points to a
-                # directory.
-                # https://docs.python.org/3.6/library/zipfile.html#zipfile.ZipInfo.is_dir
-                if not zipinfo.filename.endswith("/"):
-                    default_storage.save(
-                        os.path.join(self.extract_folder_path, zipinfo.filename),
-                        scorm_zipfile.open(zipinfo.filename),
-                    )
+        # Clean storage folder, if it already exists
+        self.clean_storage()
 
+        # Then, extract zip file
         try:
+            self.extract_package(package_file)
             self.update_package_fields()
         except ScormError as e:
             response["errors"].append(e.args[0])
 
         return self.json_response(response)
 
+    def clean_storage(self):
+        if self.storage.exists(self.extract_folder_base_path):
+            logger.info(
+                'Removing previously unzipped "%s"', self.extract_folder_base_path
+            )
+            self.recursive_delete(self.extract_folder_base_path)
+
+    def recursive_delete(self, root):
+        """
+        Recursively delete the contents of a directory in the Django default storage.
+        Unfortunately, this will not delete empty folders, as the default FileSystemStorage
+        implementation does not allow it.
+        """
+        directories, files = self.storage.listdir(root)
+        for directory in directories:
+            self.recursive_delete(os.path.join(root, directory))
+        for f in files:
+            self.storage.delete(os.path.join(root, f))
+
+    def extract_package(self, package_file):
+        with zipfile.ZipFile(package_file, "r") as scorm_zipfile:
+            zipinfos = scorm_zipfile.infolist()
+            root_path = None
+            root_depth = -1
+            # Find root folder which contains imsmanifest.xml
+            for zipinfo in zipinfos:
+                if os.path.basename(zipinfo.filename) == "imsmanifest.xml":
+                    depth = len(os.path.split(zipinfo.filename))
+                    if depth < root_depth or root_depth < 0:
+                        root_path = os.path.dirname(zipinfo.filename)
+                        root_depth = depth
+
+            if root_path is None:
+                raise ScormError(
+                    "Could not find 'imsmanifest.xml' file in the scorm package"
+                )
+
+            for zipinfo in zipinfos:
+                # Extract only files that are below the root
+                if zipinfo.filename.startswith(root_path):
+                    # Do not unzip folders, only files. In Python 3.6 we will have access to
+                    # the is_dir() method to verify whether a ZipInfo object points to a
+                    # directory.
+                    # https://docs.python.org/3.6/library/zipfile.html#zipfile.ZipInfo.is_dir
+                    if not zipinfo.filename.endswith("/"):
+                        dest_path = os.path.join(
+                            self.extract_folder_path,
+                            os.path.relpath(zipinfo.filename, root_path),
+                        )
+                        self.storage.save(
+                            dest_path,
+                            scorm_zipfile.open(zipinfo.filename),
+                        )
+
     @property
     def index_page_url(self):
         if not self.package_meta or not self.index_page_path:
             return ""
         folder = self.extract_folder_path
-        if default_storage.exists(
+        if self.storage.exists(
             os.path.join(self.extract_folder_base_path, self.index_page_path)
         ):
             # For backward-compatibility, we must handle the case when the xblock data
             # is stored in the base folder.
             folder = self.extract_folder_base_path
             logger.warning("Serving SCORM content from old-style path: %s", folder)
-        return default_storage.url(os.path.join(folder, self.index_page_path))
+        return self.storage.url(os.path.join(folder, self.index_page_path))
 
     @property
     def package_path(self):
@@ -328,44 +391,64 @@ class ScormXBlock(XBlock):
         """
         Update version and index page path fields.
         """
-        self.index_page_path = ""
-        imsmanifest_path = os.path.join(self.extract_folder_path, "imsmanifest.xml")
-        try:
-            imsmanifest_file = default_storage.open(imsmanifest_path)
-        except IOError:
-            raise ScormError(
-                "Invalid package: could not find 'imsmanifest.xml' file at the root of the zip file"
+        self.index_page_path = self.find_relative_file_path("index.html")
+        imsmanifest_path = self.find_file_path("imsmanifest.xml")
+        imsmanifest_file = self.storage.open(imsmanifest_path)
+        tree = ET.parse(imsmanifest_file)
+        imsmanifest_file.seek(0)
+        namespace = ""
+        for _, node in ET.iterparse(imsmanifest_file, events=["start-ns"]):
+            if node[0] == "":
+                namespace = node[1]
+                break
+        root = tree.getroot()
+
+        if namespace:
+            resource = root.find("{{{0}}}resources/{{{0}}}resource".format(namespace))
+            schemaversion = root.find(
+                "{{{0}}}metadata/{{{0}}}schemaversion".format(namespace)
             )
         else:
-            tree = ET.parse(imsmanifest_file)
-            imsmanifest_file.seek(0)
-            self.index_page_path = "index.html"
-            namespace = ""
-            for _, node in ET.iterparse(imsmanifest_file, events=["start-ns"]):
-                if node[0] == "":
-                    namespace = node[1]
-                    break
-            root = tree.getroot()
+            resource = root.find("resources/resource")
+            schemaversion = root.find("metadata/schemaversion")
 
-            if namespace:
-                resource = root.find(
-                    "{{{0}}}resources/{{{0}}}resource".format(namespace)
-                )
-                schemaversion = root.find(
-                    "{{{0}}}metadata/{{{0}}}schemaversion".format(namespace)
-                )
-            else:
-                resource = root.find("resources/resource")
-                schemaversion = root.find("metadata/schemaversion")
+        if resource:
+            self.index_page_path = resource.get("href")
+        if (schemaversion is not None) and (
+            re.match("^1.2$", schemaversion.text) is None
+        ):
+            self.scorm_version = "SCORM_2004"
+        else:
+            self.scorm_version = "SCORM_12"
 
-            if resource:
-                self.index_page_path = resource.get("href")
-            if (schemaversion is not None) and (
-                re.match("^1.2$", schemaversion.text) is None
-            ):
-                self.scorm_version = "SCORM_2004"
-            else:
-                self.scorm_version = "SCORM_12"
+    def find_relative_file_path(self, filename):
+        return os.path.relpath(self.find_file_path(filename), self.extract_folder_path)
+
+    def find_file_path(self, filename):
+        """
+        Search recursively in the extracted folder for a given file. Path of the first
+        found file will be returned. Raise a ScormError if file cannot be found.
+        """
+        path = self.get_file_path(filename, self.extract_folder_path)
+        if path is None:
+            raise ScormError(
+                "Invalid package: could not find '{}' file".format(filename)
+            )
+        return path
+
+    def get_file_path(self, filename, root):
+        """
+        Same as `find_file_path`, but don't raise error on file not found.
+        """
+        subfolders, files = self.storage.listdir(root)
+        for f in files:
+            if f == filename:
+                return os.path.join(root, filename)
+        for subfolder in subfolders:
+            path = self.get_file_path(filename, os.path.join(root, subfolder))
+            if path is not None:
+                return path
+        return None
 
     def get_completion_status(self):
         completion_status = self.lesson_status
@@ -379,11 +462,7 @@ class ScormXBlock(XBlock):
         accessible at a url with that also includes this name.
         """
         default_scorm_location = "scorm"
-        settings_service = self.runtime.service(self, "settings")
-        if not settings_service:
-            return default_scorm_location
-        xblock_settings = settings_service.get_settings_bucket(self)
-        return xblock_settings.get("LOCATION", default_scorm_location)
+        return self.xblock_settings.get("LOCATION", default_scorm_location)
 
     @staticmethod
     def get_sha1(file_descriptor):
@@ -408,7 +487,7 @@ class ScormXBlock(XBlock):
         if self.index_page_url:
             return {
                 "last_modified": self.package_meta.get("last_updated", ""),
-                "scorm_data": default_storage.url(self.package_path),
+                "scorm_data": self.storage.url(self.package_path),
                 "size": self.package_meta.get("size", 0),
                 "index_page": self.index_page_path,
             }
@@ -427,18 +506,31 @@ class ScormXBlock(XBlock):
             ),
         ]
 
+    @property
+    def storage(self):
+        """
+        Return the storage backend used to store the assets of this xblock. This is a cached property.
+        """
+        if not getattr(self, "_storage", None):
+            def get_default_storage(_xblock):
+                return default_storage
+        
+            storage_func = self.xblock_settings.get("STORAGE_FUNC", get_default_storage)
+            if isinstance(storage_func, string_types):
+                storage_func = import_string(storage_func)
+            self._storage = storage_func(self)
+        
+        return self._storage
 
-def recursive_delete(root):
-    """
-    Recursively delete the contents of a directory in the Django default storage.
-    Unfortunately, this will not delete empty folders, as the default FileSystemStorage
-    implementation does not allow it.
-    """
-    directories, files = default_storage.listdir(root)
-    for directory in directories:
-        recursive_delete(os.path.join(root, directory))
-    for f in files:
-        default_storage.delete(os.path.join(root, f))
+    @property
+    def xblock_settings(self):
+        """
+        Return a dict of settings associated to this XBlock.
+        """
+        settings_service = self.runtime.service(self, "settings") or {}
+        if not settings_service:
+            return {}
+        return settings_service.get_settings_bucket(self)
 
 
 class ScormError(Exception):
